package br.usp.ime.mac5855.knn;

/**
 * Esta classe define um comparador que calcula a distncia euclideana entre
 * dois objetos, a ser usado no algoritmo Knn.
 * A distncia euclideana  a raiz quadrada da soma dos quadrados das distncias
 * em cada dimenso (vide teorema de Pitgoras).
 *
 * @param <E> O tipo de objeto a ser comparado.
 * @see Escalonador
 * @author Victor Williams Stafusa da Silva
 */
public final class ComparadorEuclideano<E> implements ComparadorKnn<E> {

    /** Objeto que sabe medir as escalas e os atributos dos elementos. */
    private final Escalonador<E> escalonador;

    /**
     * Cria um {@code ComparadorEuclideano} com base em um {@code Escalonador}
     * que  usado para medir as escalas e os atributos dos elementos.
     * @param escalonador O objeto que mede as escalas e obtm os valores dos
     * atributos dos elementos dados.
     */
    public ComparadorEuclideano(Escalonador<E> escalonador) {
        this.escalonador = escalonador;
    }

    /**
     * Calcula a distncia euclideana entre dois elementos.
     * A distncia euclideana  calculada pel teorema de Pitgoras, somando-se
     * o quadrado da distncia em cada dimenso e retirando-se a raiz quadrada.
     * @param elem1 O primeiro elemento.
     * @param elem2 O segundo elemento.
     * @return A distncia entre os dois elementos, que sempre  maior ou igual
     * a zero.
     * @throws IllegalArgumentException Se qualquer um dos parmetros for
     * {@code null} ou se no for possvel medir a distncia entre eles.
     */
    @Override
    public double distancia(E elem1, E elem2) {

        // Obtm os valores dos atributos dos objetos.
        double[] valores1 = escalonador.valores(elem1);
        double[] valores2 = escalonador.valores(elem2);

        // Verifica se os valores fazem sentido.

        if (valores1 == null || valores2 == null) {
            throw new IllegalArgumentException("A valorao dos elementos no  vlida. No pode ser null.");
        }

        if (valores1.length == 0 || valores2.length == 0) {
            throw new IllegalArgumentException("A valorao dos elementos no  vlida. O nmero de dimenses no pode ser zero.");
        }

        if (valores1.length != valores2.length) {
            throw new IllegalArgumentException("A valorao dos elementos no  vlida. O nmero de dimenses deve ser igual para todos os elementos.");
        }

        double distancia = 0.0;

        // Soma os quadrados das distncias de cada dimenso.
        for (int i = 0; i < valores1.length; i++) {
            double valor = escalonador.peso(i, valores1[i] - valores2[i]);
            distancia += valor * valor;
        }

        // Tira a raiz quadrada e pronto, a distncia est calculada.
        return Math.sqrt(distancia);
    }
}