package br.usp.ime.mac5855.knn;

/**
 * Esta classe representa uma implementao parcial de escalonador, que escalona
 * os atributos multiplicando-o cada um por um peso pr-determinado. Apenas o
 * mtodo {@linkplain Escalonador#valores(Object) Escalonador.valores(E)} deve
 * ser implementado nas subclasses.
 * @param <E> O tipo de objeto do qual os atributos so avaliados.
 * @author Victor Williams Stafusa da Silva
 */
public abstract class EscalonadorMultiplicativo<E> implements Escalonador<E> {

    /**
     * Representa os pesos de cada atributo. Embora a referncia ao array seja
     * final, os valores do array podem ser alterados.
     */
    private final double[] pesos;

    /**
     * Cria uma instncia informando os pesos dados para cada atributo.
     * @param pesos Array representando os pesos de cada atributo. Se este
     * array for modificado posteriormente os pesos sero alterados.
     * @throws IllegalArgumentException Se o array de pesos for {@code null} ou
     * tiver tamanho zero.
     */
    public EscalonadorMultiplicativo(double[] pesos) {

        // Valida se o array tem sentido.
        if (pesos == null) {
            throw new IllegalArgumentException("O array de dimenses no pode ser null.");
        }
        if (pesos.length == 0) {
            throw new IllegalArgumentException("O nmero de dimenses deve ser pelo menos 1.");
        }
        this.pesos = pesos;
    }

    /**
     * Obtm os pesos de cada atributo. Estes pesos podem ser alterados ao
     * alterar-se os valores do array retornado. Corresponde ao exato mesmo
     * array que  informado no construtor.
     * @return Os pesos de cada atributo.
     */
    protected double[] getPesos() {
        return pesos;
    }

    /**
     * Escalona um determinado atributo, multiplicando-o pelo peso
     * correspondente.
     * @param numDimensao O ndice da dimenso correspondente a um atributo, de
     * acordo com o array retornado pelo mtodo
     * {@linkplain Escalonador#valores(Object) Escalonador.valores(E)}.
     * @param valorOriginal O valor a ser escalonado.
     * @return O valor do atributo escalonado, correpondente ao produto do
     * parmetro {@code valorOriginal} pelo peso correspondente.
     */
    @Override
    public double peso(int numDimensao, double valorOriginal) {
        return valorOriginal * pesos[numDimensao];
    }
}