package br.usp.ime.mac5855.knn;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Esta classe implementa o algoritmo Knn. Para utiliz-la, basta criar uma
 * instncia informando uma estratgia adequada e popul-la com uma relao de
 * preos dos elementos. Depois de populada, basta utilizar os mtodos
 * {@linkplain #estimarPreco(Object) estimarPreco(E)} ou
 * {@linkplain #probabilidadePrecoFaixa(Object, double, double)
 * probabilidadePrecoFaixa(E, double, double)} para fazer-se predies.
 * Novos elementos podem ser adicionados ou removidos posteriormente por meio
 * dos mtodos {@linkplain #add(Object, double) add(E, double)},
 * {@linkplain #addAll(Map) addAll(Map&lt;E, Double&gt;)} e
 * {@linkplain #remove(Object) remove(E)}.
 * O preo de um item pode ser atualizado ao adicion-lo novamente com um novo
 * preo. Um mesmo item nunca  armazenado mais do que uma vez em uma mesma
 * instncia desta classe.
 *
 * @param <E> O tipo de objeto a ser avaliado.
 * @see KnnStrategy
 *
 * @author Victor Williams Stafusa da Silva
 */
public final class KnnPonderado<E> {

    /** Relao de preos dos produtos conhecidos. */
    private final Map<E, Double> elementosConhecidosComPreco;

    /** Estratgia usada para fazer-se medies, ponderaes, etc. */
    private final KnnStrategy<E> strategy;

    /**
     * Cria uma instncia do algoritmo Knn no populada.
     * @param strategy A estratgia a ser usada para fazer-se predies,
     * ponderaes, etc.
     * @throws IllegalArgumentException Se a estratgia for {@code null}.
     */
    public KnnPonderado(KnnStrategy<E> strategy) {
        this(null, strategy);
    }

    /**
     * Cria uma instncia do algoritmo Knn j populada.
     * @param elementos Relao de preos inicial dos elementos.
     * @param strategy A estratgia a ser usada para fazer-se predies,
     * ponderaes, etc.
     * @throws IllegalArgumentException Se a estratgia for {@code null}.
     */
    public KnnPonderado(Map<E, Double> elementos, KnnStrategy<E> strategy) {

        // Valida e define a estratgia.
        if (strategy == null) {
            throw new IllegalArgumentException("O strategy no pode ser nulo.");
        }
        this.strategy = strategy;

        // Cria a populao inicial de elementos.
        this.elementosConhecidosComPreco = new HashMap<E, Double>();
        if (elementos != null) {
            addAll(elementos);
        }
    }

    /**
     * Adiciona um elemento a lista de preos.
     * Caso o elemento j exista na lista de preos, o preo dele  atualizado.
     * @param elemento O elemento a ser adicionado.
     * @param preco O preo do elemento.
     * @throws IllegalArgumentException Se o elemento a ser adicionado for
     * {@code null}.
     */
    public void add(E elemento, double preco) {
        if (elemento == null) {
            throw new IllegalArgumentException("No  possvel adicionar-se um elemento null.");
        }
        elementosConhecidosComPreco.put(elemento, preco);
    }

    /**
     * Adiciona um elemento a lista de preos.
     * @param elementos Uma relao dos elementos e dos respectivos preos a ser
     * adicionada.
     * @throws IllegalArgumentException Se a relao de elementos a ser
     * adicionada for {@code null}.
     */
    public void addAll(Map<E, Double> elementos) {
        if (elementos == null) {
            throw new IllegalArgumentException("A lista de preos no pode ser null.");
        }
        for (Map.Entry<E, Double> entry : elementos.entrySet()) {
            this.elementosConhecidosComPreco.put(entry.getKey(), entry.getValue());
        }
    }

    /**
     * Remove um elemento da lista de preos.
     * Caso o elemento no exista na lista de preos ou se este for
     * {@code null}, ento nada acontece.
     * @param elemento O elemento a ser removido.
     */
    public void remove(E elemento) {
        if (elemento == null) return;
        elementosConhecidosComPreco.remove(elemento);
    }

    /**
     * Calcula os vizinhos e suas distncias para um dado elemento.  retornado
     * uma relao das distncias para uma lista de elementos. Normalmente tal
     * lista conter apenas um elemento, mas em caso de empates (dois ou mais
     * elementos a uma mesma distncia do elemento dado), esta lista conter
     * mais de um elemento. A relao  ordenada com os vizinhos mais prximos
     * primeiro.
     * @param elemento O elemento do qual deseja-se obter os vizinhos.
     * @return Mapa relacionando cada distncia a uma lista de vizinhos.
     */
    private Map<Double, List<E>> vizinhosPorDistancia(E elemento) {

        // Cria uma relao de vizinhos.
        Map<Double, List<E>> vizinhos = new TreeMap<Double, List<E>>();

        // Popula a relao com todos os vizinhos.
        for (E conhecido : elementosConhecidosComPreco.keySet()) {

            // Calcula a distncia.
            double distancia = strategy.getComparador().distancia(elemento, conhecido);

            // Adiciona o elemento a lista de vizinhos na distncia encontrada.
            List<E> lista = vizinhos.get(distancia);
            if (lista == null) {
                lista = new ArrayList<E>();
            }
            lista.add(conhecido);
            vizinhos.put(distancia, lista);
        }

        return vizinhos;
    }

    /**
     * D uma estimativa de preo para um objeto.
     * @param elemento O objeto a ter o preo estimado.
     * @return O preo estimado. No caso de no ser possvel estimar-se o preo,
     * zero  retornado.
     * @throws IllegalArgumentException Se o elemento a ter o preo estimado for
     * {@code null}.
     */
    public double estimarPreco(E elemento) {

        if (elemento == null) {
            throw new IllegalArgumentException("No  possvel estimar-se o preo de um elemento null.");
        }

        // Relao de elementos e suas respectivas distncias.
        Map<Double, List<E>> vizinhos = vizinhosPorDistancia(elemento);

        int iteracoesFaltando = strategy.getNumeroVizinhos();
        double estimativa = 0.0;
        double pesoTotal = 0.0;

        // Itera pelos k vizinhos mais prximos.
        a: for (Map.Entry<Double, List<E>> entry : vizinhos.entrySet()) {
            for (E elem : entry.getValue()) {

                // Obtm o preo do vizinho e o seu peso.
                double peso = strategy.getPonderador().ponderar(entry.getKey());
                pesoTotal += peso;
                estimativa += elementosConhecidosComPreco.get(elem) * peso;

                // Conta as iteraes.
                iteracoesFaltando--;
                if (iteracoesFaltando == 0) break a;
            }
        }

        // Devolve a estimativa do preo normalizada, ou zero se no foi
        // possvel fazer-se uma predio.
        return pesoTotal == 0.0 ? 0.0 : estimativa / pesoTotal;
    }

    /**
     * Calcula a probabilidade (entre 0.0 e 1.0) de um objeto ter um preo
     * dentro de uma determinada faixa.
     * @param elemento O objeto do qual deseja-se calcular a probabilidade do
     * preo estar na faixa determinada.
     * @param precoMinimo Incio da faixa de preos.
     * @param precoMaximo Fim da faixa de preos.
     * @return A probabilidade do preo do elemento dado estar dentro da faixa
     * especificada, em uma escala de 0.0 a 1.0.
     * @throws IllegalArgumentException Se o elemento determinado for
     * {@code null} ou se o incio da faixa de preos for maior do que o fim.
     */
    public double probabilidadePrecoFaixa(E elemento, double precoMinimo, double precoMaximo) {

        if (elemento == null) {
            throw new IllegalArgumentException("No  possvel calcular-se probabilidade de preo de um elemento null.");
        }
        if (precoMinimo > precoMaximo) {
            throw new IllegalArgumentException("Faixa de preos invlida.");
        }

        // Relao de elementos e suas respectivas distncias.
        Map<Double, List<E>> vizinhos = vizinhosPorDistancia(elemento);

        int iteracoesFaltando = strategy.getNumeroVizinhos();
        double pesoTotal = 0.0;
        double pesoTotalNaFaixa = 0.0;

        // Itera pelos k vizinhos mais prximos.
        a: for (Map.Entry<Double, List<E>> entry : vizinhos.entrySet()) {
            for (E elem : entry.getValue()) {

                // Obtm o preo do vizinho e o seu peso.
                double peso = strategy.getPonderador().ponderar(entry.getKey());
                pesoTotal += peso;
                double preco = elementosConhecidosComPreco.get(elem);

                // Verifica se o preo do vizinho est na faixa determinada.
                if (precoMinimo <= preco && preco <= precoMaximo) {
                    pesoTotalNaFaixa += peso;
                }

                // Conta as iteraes.
                iteracoesFaltando--;
                if (iteracoesFaltando == 0) break a;
            }
        }

        // Devolve a probabilidade.
        return pesoTotal == 0.0 ? 0.0 : pesoTotalNaFaixa / pesoTotal;
    }
}
