package br.usp.ime.mac5855.knn;

/**
 * Ponderado que pondera os valores por meio de uma curva gaussiana (curva
 * sino). Esta curva d um valor maior para itens prximos (mas sem que este
 * valor tenda ao infinito) e valores prximos de zero para itens distantes
 * (mas, nunca igual a zero). A curvatura em si  contralada por meio de dois
 * parmetros: o sigma (desvio padro) e o multiplicador (valor constante
 * multiplicado  funo).
 * @author Victor Williams Stafusa da Silva
 */
public final class PonderadorGaussiano implements PonderadorKnn {

    /**
     * O valor padro para o sigma.
     */
    private static final double SIGMA_DEFAULT = 10.0;

    /**
     * O valor padro para o multiplicador.
     */
    private static final double MULTIPLICADOR_DEFAULT = 1.0;

    /**
     * O denominador do expoente na funo gaussiana, que  obtido a partir do
     * sigma.
     */
    private final double denominadorExpoente;

    /**
     * Valor constante a ser multiplicado pelo resultado na ponderao.
     */
    private final double multiplicador;

    /**
     * Cria um {@code PonderadorGaussiano} com os valores padro para o sigma e
     * o multiplicador, valores estes que so 10.0 e 1.0, respectivamente.
     */
    public PonderadorGaussiano() {
        this(SIGMA_DEFAULT);
    }

    /**
     * Cria um {@code PonderadorGaussiano} com o valor especificado para o sigma
     * o com o valor padro para o multiplicador (que  1.0).
     * @param sigma O valor do sigma (desvio padro).
     */
    public PonderadorGaussiano(double sigma) {
        this(MULTIPLICADOR_DEFAULT, sigma);
    }

    /**
     * Cria um {@code PonderadorGaussiano} com o valor especificado para o sigma
     * o com o valor padro para o multiplicador.
     * @param multiplicador O valor a ser multiplicado no resultado da
     * ponderao.
     * @param sigma O valor do sigma (desvio padro).
     */
    public PonderadorGaussiano(double multiplicador, double sigma) {
        if (multiplicador <= 0.0) throw new IllegalArgumentException("O multiplicador no pode ser menor ou igual a zero.");
        if (sigma <= 0.0) throw new IllegalArgumentException("O sigma no pode ser menor ou igual a zero.");
        this.denominadorExpoente = sigma * sigma * 2.0;
        this.multiplicador = multiplicador;
    }

    /**
     * Pondera um valor representando uma distncia, de acordo com uma funo
     * gaussiana (curva sino), que fornece o valor do resultado.
     * @param valor O valor a ser ponderado.
     * @return O valor aps a ponderao.
     */
    @Override
    public double ponderar(double valor) {
        return multiplicador * Math.exp((-valor * valor) / denominadorExpoente);
    }
}