package br.usp.ime.mac5855.knn;

/**
 * Ponderador que pondera os valores aplicando-lhes o inverso, ou seja,
 * f(x)=1/x. Na verdade, para evitar o prolema de diviso por zero, e para
 * se regular melhor a escala, a funo usada  f(x)=N/(x+C), sendo que tanto o
 * valor de N (numerador) e de C (constante do denominador) podem ser definidos.
 * O valor padro para o numerador  1, e para a constante do denominador  0.1.
 * @author Victor Williams Stafusa da Silva
 */
public final class PonderadorInverso implements PonderadorKnn {

    /**
     * A constante do denominador a ser usada como o valor C na frmula
     * f(x)=N/(x+c) usada por este ponderador.
     */
    private final double constanteDenominador;

    /**
     * O numerador a ser usado como o valor N na frmula f(x)=N/(x+c) usada por
     * este ponderador.
     */
    private final double numerador;

    /** Define o valor padro do numerador. */
    private static final double NUMERADOR_DEFAULT = 1.0;

    /** Define o valor padro da constante do denominador. */
    private static final double CONSTANTE_DENOMINADOR_DEFAULT = 0.1;

    /**
     * Cria um {@code PonderadorInverso} utilizando os valores padro para o
     * numerador e para a constante do denominador.
     */
    public PonderadorInverso() {
        this(NUMERADOR_DEFAULT, CONSTANTE_DENOMINADOR_DEFAULT);
    }

    /**
     * Cria um {@code PonderadorInverso} utilizando os valores do numerador e da
     * constante do denominador informados.
     * @param numerador O valor do numerador informado.
     * @param constanteDenominador A constante do denominador informada.
     * @throws IllegalArgumentException Se qualquer um dos parmetros for
     * negativo.
     */
    public PonderadorInverso(double numerador, double constanteDenominador) {
        if (constanteDenominador < 0.0) {
            throw new IllegalArgumentException("O denominador no pode ser negativo.");
        }
        if (numerador < 0.0) {
            throw new IllegalArgumentException("O numerador no pode ser negativo.");
        }
        this.numerador = numerador;
        this.constanteDenominador = constanteDenominador;
    }

    /**
     * Obtm o numerador a ser usado como o valor N na frmula f(x)=N/(x+c)
     * usada por este ponderador.
     * @return O valor do numerador.
     */
    public double getNumerador() {
        return numerador;
    }

    /**
     * Obtm a constante do denominador a ser usada como o valor C na frmula
     * f(x)=N/(x+c) usada por este ponderador.
     * @return A constante do denominador.
     */
    public double getConstanteDenominador() {
        return constanteDenominador;
    }

    /**
     * Pondera um valor representando uma distncia, calculando o seu inverso.
     * Este valor tende ao infinito quanto menor for a distncia e tende a zero
     * quanto maior for a distncia.
     * @param valor O valor a ser ponderado.
     * @return O valor aps a ponderao. Aproximadamente o inverso do valor
     * dado, multiplicado por alguma constante.
     */
    @Override
    public double ponderar(double valor) {
        return numerador / (constanteDenominador + valor);
    }
}
