package br.usp.ime.mac5855.knn.validacaocruzada;

import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Especifica uma estratgia de particionamento aleatrio para os dados da
 * classe {@linkplain ValidacaoCruzada}.
 * @param <E> O tipo dos objeto a serem particionados.
 * @author Victor Williams Stafusa da Silva
 */
public final class ParticionadorRandom<E> implements Particionador<E> {

    /** Gerador de nmeros aleatrios usado por esta classe. */
    private Random random;

    /** O conjunto de treinamento atual. */
    private Set<E> treinamento;

    /** O conjunto de teste atual. */
    private Set<E> teste;

    /**
     * A proporo de elementos a serem colocados no conjunto de teste, em uma
     * escala de 0.0 a 1.0.
     */
    private double proporcaoTeste;

    /**
     * Cria um particionador aleatrio, especificando qual  a proporo de
     * dados no conjunto de testes, em uma escala de 0.0 a 1.0.
     * @param proporcaoTeste A proporo de elementos a serem colocados no
     * conjunto de teste.
     */
    public ParticionadorRandom(double proporcaoTeste) {
        this(proporcaoTeste, new Random());
    }

    /**
     * Cria um particionador aleatrio, especificando qual  a proporo de
     * dados no conjunto de testes, em uma escala de 0.0 a 1.0 e qual  o
     * nmero a ser usado como semente para criar um gerador de nmeros
     * aleatrios deste objeto.
     * @param proporcaoTeste A proporo de elementos a serem colocados no
     * conjunto de teste.
     * @param seed A semente usada para criar-se o gerador de nmero aleatrios
     * deste objeto.
     */
    public ParticionadorRandom(double proporcaoTeste, long seed) {
        this(proporcaoTeste, new Random(seed));
    }

    /**
     * Cria um particionador aleatrio, especificando qual  a proporo de
     * dados no conjunto de testes, em uma escala de 0.0 a 1.0 e um gerador de
     * nmeros aleatrios a ser usado por este objeto.
     * @param proporcaoTeste A proporo de elementos a serem colocados no
     * conjunto de teste.
     * @param random O gerador de nmero aleatrios deste objeto.
     */
    public ParticionadorRandom(double proporcaoTeste, Random random) {
        this.setGerador(random);
        this.setProporcaoTeste(proporcaoTeste);
    }

    /**
     * Obtm o gerador de nmeros aleatrios usado por este objeto.
     * @return O gerador de nmeros aleatrios usado.
     */
    public Random getGerador() {
        return this.random;
    }

    /**
     * Define o gerador de nmeros aleatrios a ser usado por este objeto.
     * @param random O gerador de nmeros aleatrios a ser usado.
     * @throws IllegalArgumentException Se o parmetro informado for
     * {@code null}.
     */
    public void setGerador(Random random) {
        if (random == null) {
            throw new IllegalArgumentException("O gerador de nmeros aleatrios no pode ser null.");
        }
        this.random = random;
    }

    /**
     * Obtm a proporo da populao que ser utilizada como dados de teste,
     * em uma escala de 0.0 a 1.0.
     * @return A proporo da populao a ser usada como dados de teste.
     */
    public double getProporcaoTeste() {
        return this.proporcaoTeste;
    }

    /**
     * Define a proporo da populao que ser utilizada como dados de teste,
     * em uma escala de 0.0 a 1.0.
     * @param proporcaoTeste A proporo da populao a ser usada como dados de
     * teste.
     * @throws IllegalArgumentException Se o parmetro informado estiver fora da
     * faixa vlida.
     */
    public void setProporcaoTeste(double proporcaoTeste) {
        if (proporcaoTeste < 0.0 || proporcaoTeste > 1.0) {
            throw new IllegalArgumentException("A proporo dos dados de teste deve estar entre 0.0 e 1.0.");
        }
        this.proporcaoTeste = proporcaoTeste;
    }

    /**
     * Particiona aleatoriamente uma populao de dados em conjuntos de
     * treinamento e de teste, a serem obtidos posteriormente pelos mtodos
     * {@linkplain #getConjuntoTreinamento()} e
     * {@linkplain #getConjuntoTeste()}.
     * @param populacao {@inheritDoc}
     * @throws IllegalArgumentException {@inheritDoc}
     */
    @Override
    public void particionar(Set<E> populacao) {

        // Valida o parmero.
        if (populacao == null) {
            throw new IllegalArgumentException("A populao informada no pode ser nula.");
        }

        // Calcula o tamanho de cada um dos conjuntos a ser criado.
        int tamanhoPopulacao = populacao.size();
        int tamanhoTeste = (int) (tamanhoPopulacao * proporcaoTeste);
        int tamanhoTreinamento = tamanhoPopulacao - tamanhoTeste;

        // Cria o conjunto de treinamento e de teste, inicialmente vazios.
        this.treinamento = new HashSet<E>();
        this.teste = new HashSet<E>();

        // Percorre a populao colocando os elementos, ora no conjunto de
        // treinamento, ora no de teste. Contabiliza a quantidade que deve ser
        // colocada no treinamento e a quantidade no teste para a prxima
        // iterao.
        for (E elemento : populacao) {
            if (random.nextInt(tamanhoPopulacao) < tamanhoTeste) {
                teste.add(elemento);
                tamanhoTeste--; // Um elemento a menos para colocar.
            } else {
                treinamento.add(elemento);
                tamanhoTreinamento--; // Um elemento a menos para colocar.
            }
            tamanhoPopulacao--; // Um elemento a menos para contabilizar.
        }
    }

    /**
     * {@inheritDoc}
     * @return {@inheritDoc}
     * @throws IllegalStateException {@inheritDoc}
     */
    @Override
    public Set<E> getConjuntoTreinamento() {
        if (treinamento == null) {
            throw new IllegalStateException("Os dados ainda no foram particionados.");
        }
        return treinamento;
    }

    /**
     * {@inheritDoc}
     * @return {@inheritDoc}
     * @throws IllegalStateException {@inheritDoc}
     */
    @Override
    public Set<E> getConjuntoTeste() {
        if (teste == null) {
            throw new IllegalStateException("Os dados ainda no foram particionados.");
        }
        return teste;
    }
}
