package br.usp.ime.mac5855.knn.validacaocruzada;

import br.usp.ime.mac5855.knn.KnnPonderado;
import br.usp.ime.mac5855.knn.KnnStrategy;
import java.util.HashMap;
import java.util.Map;

/**
 * Esta classe implementa a validao cruzada para o algoritmo Knn. Para
 * utiliz-la, basta criar uma instncia informando as estratgias adequada para
 * dividir os dados nos conjuntos de teste e de treinamento e para computar o
 * erro acumulado. Aps criada a instncia, basta usar o mtodo
 * {@linkplain #validar(Map, int, KnnStrategy)}.
 *
 * @param <E> O tipo de objeto a ser avaliado no algoritmo Knn.
 * @see KnnPonderado
 * @see KnnStrategy
 * @see Particionador
 * @see AvaliadorErro
 * @author Victor Williams Stafusa da Silva
 */
public class ValidacaoCruzada<E> {

    /** Estratgia usada para contabilizar os erros. */
    private final AvaliadorErro avaliador;

    /**
     * Estratgia usada para dividir os dados no conjunto de teste e de
     * treinamento.
     */
    private final Particionador<E> particionador;

    /**
     * Construtor nico.
     * @param particionador A estatgia a ser usada para particionar os dados
     * nos conjuntos de treinamento e de teste.
     * @param avaliador A estatgia a ser usada para contabilizar os erros.
     */
    public ValidacaoCruzada(Particionador<E> particionador, AvaliadorErro avaliador) {
        this.avaliador = avaliador;
        this.particionador = particionador;
    }

    /**
     * Faz a validao cruzada do algoritmo Knn.
     * @param populacao Os dados a serem usados para popular o algoritmo Knn, a
     * serem particionados de forma diferente a cada tentativa.
     * @param tentativas O nmero de iteraes a serem feitas. A cada iterao,
     * os dados so particionados de uma forma diferente por meio do mtodo
     * {@linkplain Particionador#particionar(Set)} do
     * {@linkplain Particionador} informado no construtor.
     * @param knnStrategy A estratgia a ser usada pela classe
     * {@linkplain KnnPonderado}.
     * @return Os erros acumulados.
     * @throws IllegalArgumentException Se a populao ou a estratgia informada
     * for {@code null} ou se o nmero de tentativas for zero ou negativo.
     */
    public double validar(Map<E, Double> populacao, int tentativas, KnnStrategy<E> knnStrategy) {

        // Valida os parmetros.
        if (populacao == null) {
            throw new IllegalArgumentException("A populao informada no pode ser nula.");
        }
        if (knnStrategy == null) {
            throw new IllegalArgumentException("A estartgia para o algoritmo Knn no pode ser nula.");
        }
        if (tentativas <= 0) {
            throw new IllegalArgumentException("O nmero de tentativas deve ser pelo menos 1.");
        }

        // Faz vria tentativas.
        double erro = 0.0;
        for (int i = 0; i < tentativas; i++) {

            // Em cada iterao (tentativa), alimenta o particionador...
            particionador.particionar(populacao.keySet());

            // ... e cria uma instncia do KnnPonderado usando o conjunto de
            // treinamento gerado.
            Map<E, Double> treinamento = new HashMap<E, Double>();
            for (E elemento : particionador.getConjuntoTreinamento()) {
                treinamento.put(elemento, populacao.get(elemento));
            }
            KnnPonderado<E> knn = new KnnPonderado<E>(treinamento, knnStrategy);

            // Percorre o conjunto de teste no algoritmo Knn, contabilizando o
            // erro na estimativa.
            for (E elemento : particionador.getConjuntoTeste()) {
                double precoEstimado = knn.estimarPreco(elemento);
                erro += avaliador.avaliarErro(precoEstimado, populacao.get(elemento));
            }
        }

        // Retorna o erro obtido, normalizado.
        return erro / tentativas;
    }
}