package br.usp.ime.mac5855.testeknn;

import br.usp.ime.mac5855.knn.ComparadorEuclideano;
import br.usp.ime.mac5855.knn.ComparadorKnn;
import br.usp.ime.mac5855.knn.Escalonador;
import br.usp.ime.mac5855.knn.KnnPonderado;
import br.usp.ime.mac5855.knn.KnnStrategy;
import br.usp.ime.mac5855.knn.PonderadorGaussiano;
import br.usp.ime.mac5855.knn.PonderadorInverso;
import br.usp.ime.mac5855.knn.PonderadorKnn;
import br.usp.ime.mac5855.knn.PonderadorMediaSimples;
import br.usp.ime.mac5855.knn.PonderadorSubtrativo;
import br.usp.ime.mac5855.knn.validacaocruzada.ValidacaoCruzada;
import br.usp.ime.mac5855.knn.validacaocruzada.AvaliadorErroQuadratico;
import br.usp.ime.mac5855.knn.validacaocruzada.ParticionadorRandom;
import java.util.List;
import java.util.Map;
import java.util.Random;

/**
 * @author Victor Williams Stafusa da Silva
 */
class Testes {

    private Testes() {}

    private static final Escalonador<Pizza> UNIFORME = new EscalonadorPizzaUniforme();

    private static final PonderadorKnn SIMPLES = new PonderadorMediaSimples();
    private static final PonderadorKnn SUBTRATIVO = new PonderadorSubtrativo();
    private static final PonderadorKnn INVERSO = new PonderadorInverso();
    private static final PonderadorKnn GAUSSIANO = new PonderadorGaussiano(16.0);

    private static final ComparadorKnn<Pizza> EUCLIDEANO_UNIFORME = new ComparadorEuclideano<Pizza>(UNIFORME);

    private static Random getPizzaRandom() {
        return new Random(-896756475767L);
    }

    private static long VALIDACAO_CRUZADA_RANDOM_SEED = 7874875974385L;

    private static class StrategyNaoPonderado implements KnnStrategy<Pizza> {
        public ComparadorKnn<Pizza> getComparador() { return EUCLIDEANO_UNIFORME; }

        public PonderadorKnn getPonderador() { return SIMPLES; }

        public int getNumeroVizinhos() { return 3; }
    }

    private static class StrategyPonderado implements KnnStrategy<Pizza> {
        public ComparadorKnn<Pizza> getComparador() { return EUCLIDEANO_UNIFORME; }

        public PonderadorKnn getPonderador() { return GAUSSIANO; }

        public int getNumeroVizinhos() { return 10; }
    }

    private static class StrategyPonderadoComEscala implements KnnStrategy<Pizza> {
        private final ComparadorKnn<Pizza> comparador;

        public StrategyPonderadoComEscala(EscalonadorPizzaMultiplicativo escala) {
            this.comparador = new ComparadorEuclideano<Pizza>(escala);
        }

        public ComparadorKnn<Pizza> getComparador() { return comparador; }

        public PonderadorKnn getPonderador() { return GAUSSIANO; }

        public int getNumeroVizinhos() { return 10; }
    }

    private static void estimativaPreco(KnnStrategy<Pizza> strategy) {
        RandomPizzaFactory rpf = new RandomPizzaFactory(getPizzaRandom(), false);
        Map<Pizza, Double> pizzas = rpf.obterPizzasComPreco(1000);
        for (Pizza p : pizzas.keySet()) {
            System.out.println(p);
        }
        Pizza descobrir = rpf.obterPizza();

        KnnPonderado<Pizza> knn = new KnnPonderado<Pizza>(pizzas, strategy);
        double precoEstimado = knn.estimarPreco(descobrir);

        System.out.println("Preo estimado de R$ " + precoEstimado + " para " + descobrir.toString());
        System.out.println("Total de 1000 pizzas.");
    }

    public static void main(String[] args) {
        //testeCriarPizza();
        //testeEstimativaPrecoNaoPonderado();
        //testeEstimativaPrecoPonderado();
        //testeEstimativaPrecoPonderadoComEscalaEVariavelIrrelevante();
        //testeEstimativaPrecoPonderadoComEscala();
        //testeValidacaoCruzada();
        //testeProbabilidade();
        //distribuicaoProbabilidade();
    }

    public static void testeCriarPizza() {
        RandomPizzaFactory rpf = new RandomPizzaFactory(getPizzaRandom(), false);
        List<Pizza> pizzas = rpf.obterPizzas(20);
        for (Pizza p : pizzas) {
            System.out.println(p);
        }
    }

    public static void testeEstimativaPrecoNaoPonderado() {
        estimativaPreco(new StrategyNaoPonderado());
    }

    public static void testeEstimativaPrecoPonderado() {
        estimativaPreco(new StrategyPonderado());
    }

    public static void testeEstimativaPrecoPonderadoComEscalaEVariavelIrrelevante() {
        estimativaPreco(new StrategyPonderadoComEscala(EscalonadorPizzaMultiplicativo.comTamanhoSapato()));
    }

    public static void testeEstimativaPrecoPonderadoComEscala() {
        estimativaPreco(new StrategyPonderadoComEscala(EscalonadorPizzaMultiplicativo.ok()));
    }

    public static void testeValidacaoCruzada() {
        RandomPizzaFactory rpf = new RandomPizzaFactory(getPizzaRandom(), false);
        System.out.println("Criando pizzas");
        Map<Pizza, Double> pizzas = rpf.obterPizzasComPreco(1000);
        System.out.println("Pizzas criadas");

        ValidacaoCruzada<Pizza> vc = new ValidacaoCruzada<Pizza>(new ParticionadorRandom<Pizza>(0.1, VALIDACAO_CRUZADA_RANDOM_SEED), new AvaliadorErroQuadratico());
        double erro = vc.validar(pizzas, 10, new StrategyNaoPonderado());
        System.out.println("Taxa de erro, no ponderado: " + erro);

        vc = new ValidacaoCruzada<Pizza>(new ParticionadorRandom<Pizza>(0.1, VALIDACAO_CRUZADA_RANDOM_SEED), new AvaliadorErroQuadratico());
        erro = vc.validar(pizzas, 10, new StrategyPonderado());
        System.out.println("Taxa de erro, ponderado: " + erro);

        vc = new ValidacaoCruzada<Pizza>(new ParticionadorRandom<Pizza>(0.1, VALIDACAO_CRUZADA_RANDOM_SEED), new AvaliadorErroQuadratico());
        erro = vc.validar(pizzas, 10, new StrategyPonderadoComEscala(EscalonadorPizzaMultiplicativo.comTamanhoSapato()));
        System.out.println("Taxa de erro, ponderado com escala e dimenso irrelevante: " + erro);

        vc = new ValidacaoCruzada<Pizza>(new ParticionadorRandom<Pizza>(0.1, VALIDACAO_CRUZADA_RANDOM_SEED), new AvaliadorErroQuadratico());
        erro = vc.validar(pizzas, 10, new StrategyPonderadoComEscala(EscalonadorPizzaMultiplicativo.ok()));
        System.out.println("Taxa de erro, ponderado com escala: " + erro);
    }

    public static void testeProbabilidade() {
        RandomPizzaFactory rpf = new RandomPizzaFactory(getPizzaRandom(), true);
        Map<Pizza, Double> pizzas = rpf.obterPizzasComPreco(60);
        for (Pizza p : pizzas.keySet()) {
            System.out.println(p);
        }
        Pizza descobrir;
        while (true) {
            descobrir = rpf.obterPizza();
            if (descobrir.pizzaVip) break;
        }
        System.out.println("");
        System.out.println("Nossa pizza: " + descobrir.toString());

        KnnPonderado<Pizza> knn = new KnnPonderado<Pizza>(pizzas, new StrategyPonderadoComEscala(EscalonadorPizzaMultiplicativo.ok()));

        double probabilidade = knn.probabilidadePrecoFaixa(descobrir, 0.0, 15.0) * 100.0;
        System.out.println("Probabilidade de " + probabilidade + "% para R$ 0,00 a R$ 15,00.");

        probabilidade = knn.probabilidadePrecoFaixa(descobrir, 15.0, 20.0) * 100.0;
        System.out.println("Probabilidade de " + probabilidade + "% para R$ 15,00 a R$ 20,00.");

        probabilidade = knn.probabilidadePrecoFaixa(descobrir, 20.0, 25.0) * 100.0;
        System.out.println("Probabilidade de " + probabilidade + "% para R$ 20,00 a R$ 25,00.");

        probabilidade = knn.probabilidadePrecoFaixa(descobrir, 25.0, 30.0) * 100.0;
        System.out.println("Probabilidade de " + probabilidade + "% para R$ 25,00 a R$ 30,00.");

        probabilidade = knn.probabilidadePrecoFaixa(descobrir, 30.0, 40.0) * 100.0;
        System.out.println("Probabilidade de " + probabilidade + "% para R$ 30,00 a R$ 40,00.");

        probabilidade = knn.probabilidadePrecoFaixa(descobrir, 40.0, 1000.0) * 100.0;
        System.out.println("Probabilidade de " + probabilidade + "% para mais de R$ 40,00.");
    }

    public static void distribuicaoProbabilidade() {
        RandomPizzaFactory rpf = new RandomPizzaFactory(getPizzaRandom(), true);
        Map<Pizza, Double> pizzas = rpf.obterPizzasComPreco(60);
        for (Pizza p : pizzas.keySet()) {
            System.out.println(p);
        }
        Pizza descobrir;
        while (true) {
            descobrir = rpf.obterPizza();
            if (descobrir.pizzaVip) break;
        }
        System.out.println("");
        System.out.println("Nossa pizza: " + descobrir.toString());

        KnnPonderado<Pizza> knn = new KnnPonderado<Pizza>(pizzas, new StrategyPonderadoComEscala(EscalonadorPizzaMultiplicativo.ok()));

        for (int i = 0; i < 300; i++) {
            double probabilidade = knn.probabilidadePrecoFaixa(descobrir, i, i + 1) * 100.0;
            if (probabilidade == 0.0) continue;
            System.out.println("Probabilidade de " + probabilidade + "% para R$ " + i + ",00 a R$ " + (i + 1) + ",00.");
        }
    }
}